CANCELLED_APPT = "cancelled"
CLINIC = "clinic"
COMPLETE_APPT = "done"
INCOMPLETE_APPT = "incomplete"
IN_PROGRESS_APPT = "in_progress"
LATE_APPT = "late"
MISSED_APPT = "missed"
NEW_APPT = "new"
ONTIME_APPT = "ontime"
SCHEDULED_APPT = "scheduled"
UNSCHEDULED_APPT = "unscheduled"



from datetime import datetime
from decimal import Decimal
from typing import Any, Optional, Union

from edc_visit_tracking.constants import UNSCHEDULED

from ..constants import IN_PROGRESS_APPT, INCOMPLETE_APPT
from ..creators import UnscheduledAppointmentCreator
from ..models import Appointment


class AppointmentTestCaseMixin:
    def get_timepoint_from_visit_code(
        self: Any,
        instance: Any,
        visit_code: str,
    ) -> Optional[Union[float, Decimal]]:
        timepoint = None
        for v in instance.schedule.visits.timepoints:
            if v.name == visit_code:
                timepoint = v.timepoint
                break
        return timepoint

    def get_appointment(
        self,
        subject_identifier: Optional[str] = None,
        visit_code: Optional[str] = None,
        visit_code_sequence: Optional[int] = None,
        reason: Optional[str] = None,
        appt_datetime: Optional[datetime] = None,
        timepoint: Optional[Union[float, Decimal]] = None,
    ):
        if timepoint is not None:
            appointment = Appointment.objects.get(
                subject_identifier=subject_identifier,
                timepoint=timepoint,
                visit_code_sequence=visit_code_sequence,
            )
        else:
            appointment = Appointment.objects.get(
                subject_identifier=subject_identifier,
                visit_code=visit_code,
                visit_code_sequence=visit_code_sequence,
            )
        if appt_datetime:
            appointment.appt_datetime = appt_datetime
            appointment.save()
            appointment.refresh_from_db()
        if reason == UNSCHEDULED:
            appointment = self.create_unscheduled_appointment(appointment)
        appointment.appt_status = IN_PROGRESS_APPT
        appointment.save()
        appointment.refresh_from_db()
        return appointment

    @staticmethod
    def create_unscheduled_appointment(appointment):
        appointment.appt_status = INCOMPLETE_APPT
        appointment.save()
        appointment.refresh_from_db()
        appt_creator = UnscheduledAppointmentCreator(
            subject_identifier=appointment.subject_identifier,
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
            visit_code=appointment.visit_code,
            facility=appointment.facility,
        )
        return appt_creator.appointment



from django.contrib import admin
from django.utils.safestring import mark_safe
from edc_document_status.fieldsets import document_status_fieldset_tuple
from edc_document_status.modeladmin_mixins import DocumentStatusModelAdminMixin
from edc_model_admin import SimpleHistoryAdmin, audit_fieldset_tuple
from edc_model_admin.dashboard import ModelAdminSubjectDashboardMixin
from edc_visit_schedule import OnScheduleError, off_schedule_or_raise
from edc_visit_schedule.fieldsets import (
    visit_schedule_fields,
    visit_schedule_fieldset_tuple,
)

from .admin_actions import appointment_mark_as_done, appointment_mark_as_new
from .admin_site import edc_appointment_admin
from .constants import NEW_APPT
from .forms import AppointmentForm
from .models import Appointment


@admin.register(Appointment, site=edc_appointment_admin)
class AppointmentAdmin(
    ModelAdminSubjectDashboardMixin, DocumentStatusModelAdminMixin, SimpleHistoryAdmin
):

    show_cancel = True
    form = AppointmentForm
    actions = [appointment_mark_as_done, appointment_mark_as_new]
    date_hierarchy = "appt_datetime"
    list_display = (
        "subject_identifier",
        "__str__",
        "dashboard",
        "appt_datetime",
        "appt_type",
        "appt_status",
        "appt_timing",
        "schedule_name",
    )
    list_filter = ("visit_code", "appt_datetime", "appt_type", "appt_status", "appt_timing")

    additional_instructions = mark_safe(
        "To start or continue to edit FORMS for this subject, change the "
        'appointment status below to "In Progress" and click SAVE. <BR>'
        "<i>Note: You may only edit one appointment at a time. "
        "Before you move to another appointment, change the appointment "
        'status below to "Incomplete or "Done".</i>'
    )

    fieldsets = (
        (
            None,
            (
                {
                    "fields": (
                        "subject_identifier",
                        "appt_datetime",
                        "appt_type",
                        "appt_status",
                        "appt_reason",
                        "appt_timing",
                        "comment",
                    )
                }
            ),
        ),
        (
            "Timepoint",
            (
                {
                    "classes": ("collapse",),
                    "fields": (
                        "timepoint",
                        "timepoint_datetime",
                        "visit_code_sequence",
                        "facility_name",
                    ),
                }
            ),
        ),
        document_status_fieldset_tuple,
        visit_schedule_fieldset_tuple,
        audit_fieldset_tuple,
    )

    radio_fields = {
        "appt_type": admin.VERTICAL,
        "appt_status": admin.VERTICAL,
        "appt_reason": admin.VERTICAL,
        "appt_timing": admin.VERTICAL,
    }

    search_fields = ("subject_identifier",)

    def get_readonly_fields(self, request, obj=None):
        readonly_fields = super().get_readonly_fields(request, obj=obj)
        return (
            list(readonly_fields)
            + list(visit_schedule_fields)
            + [
                "subject_identifier",
                "timepoint",
                "timepoint_datetime",
                "visit_code_sequence",
                "facility_name",
            ]
        )

    def has_delete_permission(self, request, obj=None):
        """Override to remove delete permissions if OnSchedule
        and visit_code_sequence == 0.

        See `edc_visit_schedule.off_schedule_or_raise()`
        """
        has_delete_permission = super().has_delete_permission(request, obj=obj)
        if has_delete_permission and obj:
            if obj.visit_code_sequence == 0 or (
                obj.visit_code_sequence != 0 and obj.appt_status != NEW_APPT
            ):
                try:
                    off_schedule_or_raise(
                        subject_identifier=obj.subject_identifier,
                        report_datetime=obj.appt_datetime,
                        visit_schedule_name=obj.visit_schedule_name,
                        schedule_name=obj.schedule_name,
                    )
                except OnScheduleError:
                    has_delete_permission = False
        return has_delete_permission



from datetime import datetime
from decimal import Decimal

import arrow
from dateutil.relativedelta import relativedelta
from django.test import TestCase
from edc_facility.import_holidays import import_holidays
from edc_reference import site_reference_configs
from edc_utils import get_utcnow
from edc_visit_schedule.schedule import ScheduleError
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED, UNSCHEDULED

from edc_appointment.models import Appointment
from edc_appointment_app.models import SubjectVisit
from edc_appointment_app.visit_schedule import visit_schedule1, visit_schedule2

from ...constants import CANCELLED_APPT, IN_PROGRESS_APPT, INCOMPLETE_APPT, NEW_APPT
from ...creators import (
    InvalidParentAppointmentMissingVisitError,
    InvalidParentAppointmentStatusError,
    UnscheduledAppointmentCreator,
    UnscheduledAppointmentNotAllowed,
)
from ..helper import Helper


class TestUnscheduledAppointmentCreator(TestCase):

    helper_cls = Helper

    @classmethod
    def setUpClass(cls):
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        self.subject_identifier = "12345"
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule1)
        site_visit_schedules.register(visit_schedule=visit_schedule2)
        self.helper = self.helper_cls(
            subject_identifier=self.subject_identifier,
            now=arrow.Arrow.fromdatetime(datetime(2017, 1, 7), tzinfo="UTC").datetime,
        )

        site_reference_configs.register_from_visit_schedule(
            visit_models={"edc_appointment.appointment": "edc_appointment_app.subjectvisit"}
        )

    def test_unscheduled_allowed_but_raises_on_appt_status(self):
        self.helper.consent_and_put_on_schedule()
        schedule_name = "schedule1"
        visit = visit_schedule1.schedules.get(schedule_name).visits.first
        appointment = Appointment.objects.get(
            subject_identifier=self.subject_identifier,
            visit_code=visit.code,
            visit_code_sequence=0,
        )
        # subject_visit not created so expect exception because of
        # the missing subject_visit
        for appt_status in [NEW_APPT, IN_PROGRESS_APPT, CANCELLED_APPT]:
            with self.subTest(appt_status=appt_status):
                appointment.appt_status = appt_status
                appointment.save()
                self.assertEqual(appointment.appt_status, appt_status)
                self.assertRaises(
                    InvalidParentAppointmentMissingVisitError,
                    UnscheduledAppointmentCreator,
                    subject_identifier=self.subject_identifier,
                    visit_schedule_name=visit_schedule1.name,
                    schedule_name=schedule_name,
                    visit_code=visit.code,
                )
        # add a subject_visit and expect exception to be raises because
        # of appt_status
        subject_visit = SubjectVisit.objects.create(
            appointment=appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )
        appointment.refresh_from_db()
        self.assertEqual(appointment.visit, subject_visit)
        for appt_status in [NEW_APPT, IN_PROGRESS_APPT, CANCELLED_APPT]:
            with self.subTest(appt_status=appt_status):
                appointment.appt_status = appt_status
                appointment.save()
                self.assertEqual(appointment.appt_status, appt_status)
                self.assertRaises(
                    InvalidParentAppointmentStatusError,
                    UnscheduledAppointmentCreator,
                    subject_identifier=self.subject_identifier,
                    visit_schedule_name=visit_schedule1.name,
                    schedule_name=schedule_name,
                    visit_code=visit.code,
                )

    def test_unscheduled_not_allowed(self):
        self.assertRaises(
            UnscheduledAppointmentNotAllowed,
            UnscheduledAppointmentCreator,
            subject_identifier=self.subject_identifier,
            visit_schedule_name=visit_schedule2.name,
            schedule_name="schedule2",
            visit_code="5000",
        )

    def test_add_subject_visits(self):
        self.helper.consent_and_put_on_schedule()
        schedule_name = "schedule1"
        for visit in visit_schedule1.schedules.get(schedule_name).visits.values():
            with self.subTest(visit=visit):
                # get parent appointment
                appointment = Appointment.objects.get(
                    subject_identifier=self.subject_identifier,
                    visit_code=visit.code,
                    visit_code_sequence=0,
                    timepoint=visit.timepoint,
                    visit_schedule_name=visit_schedule1.name,
                    schedule_name=schedule_name,
                )
                appointment.appt_status = IN_PROGRESS_APPT
                appointment.save()
                appointment.refresh_from_db()

                # fill in subject visit report for this appointment
                subject_visit = SubjectVisit.objects.create(
                    appointment=appointment,
                    subject_identifier=self.subject_identifier,
                    report_datetime=appointment.appt_datetime,
                    reason=SCHEDULED,
                    visit_code=visit.code,
                    visit_code_sequence=0,
                    visit_schedule_name=visit_schedule1.name,
                    schedule_name=schedule_name,
                )
                appointment.refresh_from_db()
                self.assertTrue(appointment.visit, subject_visit)
                self.assertEqual(0, appointment.visit.visit_code_sequence)
                self.assertEqual(1, appointment.next_visit_code_sequence)

                # close appt (set to INCOMPLETE_APPT)
                appointment.appt_status = INCOMPLETE_APPT
                appointment.save()
                appointment.refresh_from_db()

                # create unscheduled off of this appt
                creator = UnscheduledAppointmentCreator(
                    subject_identifier=self.subject_identifier,
                    visit_schedule_name=visit_schedule1.name,
                    schedule_name=schedule_name,
                    visit_code=visit.code,
                    facility=appointment.facility,
                    appt_datetime=appointment.appt_datetime + relativedelta(days=1),
                )
                new_appointment = creator.appointment
                new_appointment.appt_status = IN_PROGRESS_APPT
                new_appointment.save()
                new_appointment.refresh_from_db()
                self.assertEqual(new_appointment.appt_status, IN_PROGRESS_APPT)

                # submit subject visit for the unscheduled appt
                subject_visit = SubjectVisit.objects.create(
                    appointment=new_appointment,
                    report_datetime=get_utcnow(),
                    reason=UNSCHEDULED,
                    visit_code=new_appointment.visit_code,
                    visit_code_sequence=new_appointment.visit_code_sequence,
                    visit_schedule_name=new_appointment.visit_schedule_name,
                    schedule_name=new_appointment.schedule_name,
                )
                self.assertEqual(1, new_appointment.visit_code_sequence)
                self.assertEqual(1, subject_visit.visit_code_sequence)

                # close the unscheduled appt (set to INCOMPLETE_APPT)
                new_appointment.appt_status = INCOMPLETE_APPT
                new_appointment.save()
                self.assertEqual(new_appointment.appt_status, INCOMPLETE_APPT)
                self.assertEqual(visit.timepoint, int(new_appointment.timepoint))

    def test_unscheduled_timepoint_not_incremented(self):
        self.helper.consent_and_put_on_schedule()
        schedule_name = "schedule1"
        visit = visit_schedule1.schedules.get(schedule_name).visits.first
        appointment = Appointment.objects.get(
            subject_identifier=self.subject_identifier, visit_code=visit.code
        )
        self.assertEqual(appointment.timepoint, Decimal("0.0"))
        SubjectVisit.objects.create(
            appointment=appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )
        appointment.appt_status = INCOMPLETE_APPT
        appointment.save()
        for index in range(1, 5):
            with self.subTest(index=index):
                creator = UnscheduledAppointmentCreator(
                    subject_identifier=appointment.subject_identifier,
                    visit_schedule_name=appointment.visit_schedule_name,
                    schedule_name=appointment.schedule_name,
                    visit_code=appointment.visit_code,
                    facility=appointment.facility,
                )
                self.assertEqual(appointment.timepoint, creator.appointment.timepoint)
                self.assertNotEqual(
                    appointment.visit_code_sequence,
                    creator.appointment.visit_code_sequence,
                )
                self.assertEqual(
                    creator.appointment.visit_code_sequence,
                    index,
                )
                SubjectVisit.objects.create(
                    appointment=creator.appointment,
                    report_datetime=get_utcnow(),
                    reason=UNSCHEDULED,
                )
                creator.appointment.appt_status = INCOMPLETE_APPT
                creator.appointment.save()

    def test_appointment_title(self):
        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.first_appointment(
            subject_identifier=self.subject_identifier,
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )
        self.assertEqual(appointment.title, "Day 1")

        SubjectVisit.objects.create(
            appointment=appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )
        appointment.appt_status = INCOMPLETE_APPT
        appointment.save()

        creator = UnscheduledAppointmentCreator(
            subject_identifier=appointment.subject_identifier,
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
            visit_code=appointment.visit_code,
            facility=appointment.facility,
        )
        self.assertEqual(creator.appointment.title, "Day 1.1")

        SubjectVisit.objects.create(
            appointment=creator.appointment, report_datetime=get_utcnow(), reason=UNSCHEDULED
        )
        creator.appointment.appt_status = INCOMPLETE_APPT
        creator.appointment.save()

        next_appointment = Appointment.objects.next_appointment(
            visit_code=appointment.visit_code,
            subject_identifier=self.subject_identifier,
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )

        SubjectVisit.objects.create(
            appointment=next_appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )
        next_appointment.appt_status = INCOMPLETE_APPT
        next_appointment.save()

        creator = UnscheduledAppointmentCreator(
            subject_identifier=next_appointment.subject_identifier,
            visit_schedule_name=next_appointment.visit_schedule_name,
            schedule_name=next_appointment.schedule_name,
            visit_code=next_appointment.visit_code,
            facility=next_appointment.facility,
        )

        self.assertEqual(creator.appointment.title, "Day 2.1")

    def test_appointment_title_if_visit_schedule_changes(self):
        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.first_appointment(
            subject_identifier=self.subject_identifier,
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )
        self.assertEqual(appointment.title, "Day 1")

        SubjectVisit.objects.create(
            appointment=appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )
        appointment.appt_status = INCOMPLETE_APPT
        appointment.save()

        next_appointment = Appointment.objects.next_appointment(
            visit_code=appointment.visit_code,
            subject_identifier=self.subject_identifier,
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )

        SubjectVisit.objects.create(
            appointment=next_appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )
        next_appointment.appt_status = INCOMPLETE_APPT
        next_appointment.visit_code = "1111"
        self.assertRaises(ScheduleError, next_appointment.save)



from typing import Optional, Type

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.db import models, transaction
from edc_facility import Facility
from edc_visit_tracking.model_mixins import VisitModelMixin
from edc_visit_tracking.stubs import SubjectVisitModelStub

from edc_appointment.stubs import AppointmentModelStub, TAppointmentModelStub


class AppointmentMethodsModelError(Exception):
    pass


class AppointmentMethodsModelMixin(models.Model):

    """Mixin of methods for the appointment model only"""

    @property
    def visit(self: AppointmentModelStub) -> SubjectVisitModelStub:
        """Returns the related visit model instance, or None"""
        visit = None
        for f in self._meta.get_fields():
            if f.related_model:
                if issubclass(f.related_model, VisitModelMixin):
                    with transaction.atomic():
                        try:
                            visit = f.related_model.objects.get(appointment=self)
                        except ObjectDoesNotExist:
                            pass
        # return getattr(self, self.related_visit_model_attr())
        return visit

    @property
    def facility(self: AppointmentModelStub) -> Facility:
        """Returns the facility instance for this facility name"""
        app_config = django_apps.get_app_config("edc_facility")
        return app_config.get_facility(name=self.facility_name)

    @classmethod
    def related_visit_model_attr(cls: TAppointmentModelStub) -> str:
        fields = []
        for f in cls._meta.get_fields():
            if f.related_model:
                if issubclass(f.related_model, VisitModelMixin):
                    fields.append(f)
                    break
        if len(fields) > 1:
            raise AppointmentMethodsModelError(
                f"More than one field on Appointment is related field to a visit model. "
                f"Got {fields}."
            )
        elif len(fields) == 0:
            raise AppointmentMethodsModelError(
                f"{cls} has no related visit model. "
                f"Expected the related visit model to be an instance "
                "of `VisitModelMixin`."
            )
        else:
            related_visit_model_attr = fields[0].name
        return related_visit_model_attr

    @classmethod
    def visit_model_cls(cls: TAppointmentModelStub) -> Type[SubjectVisitModelStub]:
        return getattr(cls, cls.related_visit_model_attr()).related.related_model

    @property
    def next_by_timepoint(self: AppointmentModelStub) -> AppointmentModelStub:
        """Returns the next appointment or None of all appointments
        for this subject for visit_code_sequence=0.
        """
        return (
            self.__class__.objects.filter(
                subject_identifier=self.subject_identifier,
                timepoint__gt=self.timepoint,
                visit_code_sequence=0,
            )
            .order_by("timepoint")
            .first()
        )

    @property
    def last_visit_code_sequence(self: AppointmentModelStub) -> Optional[int]:
        """Returns an integer, or None, that is the visit_code_sequence
        of the last appointment for this visit code that is not self.
        (ordered by visit_code_sequence).

        A sequence would be 1000.0, 1000.1, 1000.2, ...
        """
        obj = (
            self.__class__.objects.filter(
                subject_identifier=self.subject_identifier,
                visit_schedule_name=self.visit_schedule_name,
                schedule_name=self.schedule_name,
                visit_code=self.visit_code,
                visit_code_sequence__gt=self.visit_code_sequence,
            )
            .order_by("visit_code_sequence")
            .last()
        )
        if obj:
            return obj.visit_code_sequence
        return None

    @property
    def next_visit_code_sequence(self: AppointmentModelStub) -> int:
        """Returns an integer that is the next visit_code_sequence.

        A sequence would be 1000.0, 1000.1, 1000.2, ...
        """
        if self.last_visit_code_sequence:
            return self.last_visit_code_sequence + 1
        return self.visit_code_sequence + 1

    def get_last_appointment_with_visit_report(
        self: AppointmentModelStub,
    ) -> Optional[AppointmentModelStub]:
        """Returns the last appointment model instance,
        or None, with a completed visit report.

        Ordering is by appointment timepoint/visit_code_sequence
        with a completed visit report.
        """
        appointment: Optional[AppointmentModelStub] = None
        visit = (
            self.__class__.visit_model_cls()
            .objects.filter(
                appointment__subject_identifier=self.subject_identifier,
                visit_schedule_name=self.visit_schedule_name,
                schedule_name=self.schedule_name,
            )
            .order_by("appointment__timepoint", "appointment__visit_code_sequence")
            .last()
        )
        if visit:
            appointment = visit.appointment
        return appointment

    @property
    def previous_by_timepoint(self: AppointmentModelStub) -> Optional[AppointmentModelStub]:
        """Returns the previous appointment or None by timepoint
        for visit_code_sequence=0.
        """
        return (
            self.__class__.objects.filter(
                subject_identifier=self.subject_identifier,
                timepoint__lt=self.timepoint,
                visit_code_sequence=0,
            )
            .order_by("timepoint")
            .last()
        )

    @property
    def previous(self: AppointmentModelStub) -> Optional[AppointmentModelStub]:
        """Returns the previous appointment or None in this schedule
        for visit_code_sequence=0.
        """
        return self.get_previous()

    def get_previous(
        self: AppointmentModelStub, include_interim=None
    ) -> Optional[AppointmentModelStub]:
        """Returns the previous appointment model instance,
        or None, in this schedule.

        Keywords:
            * include_interim: include interim appointments
              (e.g. those where visit_code_sequence != 0)
        """
        opts = dict(
            subject_identifier=self.subject_identifier,
            visit_schedule_name=self.visit_schedule_name,
            schedule_name=self.schedule_name,
            timepoint__lt=self.timepoint,
        )
        if include_interim and self.visit_code_sequence != 0:
            opts.pop("timepoint__lt")
            opts.update(timepoint__lte=self.timepoint)
            opts.update(visit_code_sequence__gt=0)
        elif not include_interim:
            opts.update(visit_code_sequence=0)
        appointments = (
            self.__class__.objects.filter(**opts)
            .exclude(id=self.id)
            .order_by("timepoint", "visit_code_sequence")
        )
        try:
            previous_appt: Optional[AppointmentModelStub] = appointments.reverse()[0]
        except IndexError:
            previous_appt = None
        return previous_appt

    @property
    def next(self: AppointmentModelStub) -> Optional[AppointmentModelStub]:
        """Returns the next appointment or None in this schedule
        for visit_code_sequence=0.
        """
        next_appt: Optional[AppointmentModelStub] = None
        next_visit = self.schedule.visits.next(self.visit_code)
        if next_visit:
            try:
                options = dict(
                    subject_identifier=self.subject_identifier,
                    visit_schedule_name=self.visit_schedule_name,
                    schedule_name=self.schedule_name,
                    visit_code=next_visit.code,
                    visit_code_sequence=0,
                )
                next_appt = self.__class__.objects.get(**options)
            except ObjectDoesNotExist:
                pass
        return next_appt

    class Meta:
        abstract = True



# Generated by Django 2.0rc1 on 2017-11-28 10:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_appointment", "0010_auto_20171124_1120")]

    operations = [
        migrations.AlterField(
            model_name="appointment",
            name="appt_close_datetime",
            field=models.DateTimeField(
                help_text="timepoint_datetime adjusted according to the nearest available datetime for this facility",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="appointment",
            name="timepoint",
            field=models.DecimalField(
                decimal_places=1,
                help_text="timepoint from schedule",
                max_digits=6,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="appointment",
            name="timepoint_datetime",
            field=models.DateTimeField(
                help_text="Unadjusted datetime calculated from visit schedule",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="appt_close_datetime",
            field=models.DateTimeField(
                help_text="timepoint_datetime adjusted according to the nearest available datetime for this facility",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="timepoint",
            field=models.DecimalField(
                decimal_places=1,
                help_text="timepoint from schedule",
                max_digits=6,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="timepoint_datetime",
            field=models.DateTimeField(
                help_text="Unadjusted datetime calculated from visit schedule",
                null=True,
            ),
        ),
    ]



# Generated by Django 4.0.4 on 2022-05-19 20:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('edc_appointment', '0027_appointment_document_status_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='appointment',
            name='document_status_comments',
            field=models.TextField(blank=True, help_text='for example, why some data is still pending', null=True, verbose_name='Any comments related to status of this document'),
        ),
        migrations.AddField(
            model_name='historicalappointment',
            name='document_status_comments',
            field=models.TextField(blank=True, help_text='for example, why some data is still pending', null=True, verbose_name='Any comments related to status of this document'),
        ),
    ]



# Generated by Django 3.2.4 on 2021-06-23 23:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('edc_appointment', '0024_auto_20200911_0425'),
    ]

    operations = [
        migrations.AlterField(
            model_name='appointment',
            name='appt_reason',
            field=models.CharField(choices=[('scheduled', 'Routine / Scheduled'), ('unscheduled', 'Unscheduled')], help_text="The visit report's `reason for visit` will be validated against this. Refer to the protocol's documentation for the definition of a `scheduled` visit.", max_length=25, verbose_name='Reason for appointment'),
        ),
        migrations.AlterField(
            model_name='historicalappointment',
            name='appt_reason',
            field=models.CharField(choices=[('scheduled', 'Routine / Scheduled'), ('unscheduled', 'Unscheduled')], help_text="The visit report's `reason for visit` will be validated against this. Refer to the protocol's documentation for the definition of a `scheduled` visit.", max_length=25, verbose_name='Reason for appointment'),
        ),
    ]



from datetime import datetime
from typing import Any, Optional

import arrow
from django.apps import apps as django_apps
from django.db.models.deletion import ProtectedError
from edc_facility import FacilityError

from .appointment_creator import AppointmentCreator, CreateAppointmentError


class AppointmentsCreator:

    """Note: Appointments are created using this class by
    the visit schedule.

    See also: edc_visit_schedule SubjectSchedule

    """

    appointment_creator_cls = AppointmentCreator

    def __init__(
        self,
        subject_identifier: Optional[str] = None,
        visit_schedule: Optional[Any] = None,
        schedule: Optional[Any] = None,
        report_datetime: Optional[datetime] = None,
        appointment_model: Optional[Any] = None,
        skip_baseline: Optional[bool] = None,
    ):
        self.subject_identifier = subject_identifier
        self.visit_schedule = visit_schedule
        self.schedule = schedule
        self.report_datetime = report_datetime
        self.appointment_model = appointment_model
        self.skip_baseline = skip_baseline

    def create_appointments(self, base_appt_datetime=None, taken_datetimes=None):
        """Creates appointments when called by post_save signal.

        Timepoint datetimes are adjusted according to the available
        days in the facility.
        """
        app_config = django_apps.get_app_config("edc_facility")
        appointments = []
        taken_datetimes = taken_datetimes or []
        base_appt_datetime = base_appt_datetime or self.report_datetime
        base_appt_datetime = (
            arrow.Arrow.fromdatetime(base_appt_datetime, base_appt_datetime.tzinfo)
            .to("utc")
            .datetime
        )
        timepoint_dates = self.schedule.visits.timepoint_dates(dt=base_appt_datetime)
        for visit, timepoint_datetime in timepoint_dates.items():
            try:
                facility = app_config.get_facility(visit.facility_name)
            except FacilityError as e:
                raise CreateAppointmentError(
                    f"{e} See {repr(visit)}. Got facility_name={visit.facility_name}"
                )
            appointment = self.update_or_create_appointment(
                visit=visit,
                taken_datetimes=taken_datetimes,
                timepoint_datetime=timepoint_datetime,
                facility=facility,
            )
            appointments.append(appointment)
            taken_datetimes.append(appointment.appt_datetime)
        return appointments

    def update_or_create_appointment(self, **kwargs):
        """Updates or creates an appointment for this subject
        for the visit.
        """
        appointment_creator = self.appointment_creator_cls(
            subject_identifier=self.subject_identifier,
            visit_schedule_name=self.visit_schedule.name,
            schedule_name=self.schedule.name,
            appointment_model=self.appointment_model,
            skip_baseline=self.skip_baseline,
            **kwargs,
        )
        return appointment_creator.appointment

    def delete_unused_appointments(self):
        appointments = self.appointment_model.objects.filter(
            subject_identifier=self.subject_identifier,
            visit_schedule_name=self.visit_schedule.name,
            schedule_name=self.schedule.name,
        )
        for appointment in appointments:
            try:
                appointment.delete()
            except ProtectedError:
                pass
        return None

    def recalulate_from_baseline(self):
        pass



from .appointment_form_validator import (
    INVALID_APPT_DATE,
    INVALID_APPT_REASON,
    INVALID_APPT_STATUS,
    INVALID_APPT_TIMING_CRFS_EXIST,
    INVALID_APPT_TIMING_REQUISITIONS_EXIST,
    INVALID_PREVIOUS_VISIT_MISSING,
    AppointmentFormValidator,
)
from .utils import validate_appt_datetime_unique
from .window_period_form_validator_mixin import WindowPeriodFormValidatorMixin



from typing import Any, Optional

from .constants import COMPLETE_APPT, IN_PROGRESS_APPT, INCOMPLETE_APPT


class AppointmentStatusUpdaterError(Exception):
    pass


class AppointmentStatusUpdater:
    def __init__(
        self,
        appointment: Any,
        change_to_in_progress: Optional[bool] = None,
        clear_others_in_progress: Optional[bool] = None,
    ):
        self.appointment = appointment
        if not getattr(self.appointment, "id", None):
            raise AppointmentStatusUpdaterError(
                "Appointment instance must exist. Got `id` is None"
            )
        if change_to_in_progress and self.appointment.appt_status != IN_PROGRESS_APPT:
            self.appointment.appt_status = IN_PROGRESS_APPT
            self.appointment.save_base(update_fields=["appt_status"])
        if clear_others_in_progress:
            self.update_others_from_in_progress()

    def update_others_from_in_progress(self):
        opts = dict(
            visit_schedule_name=self.appointment.visit_schedule_name,
            schedule_name=self.appointment.schedule_name,
            appt_status=IN_PROGRESS_APPT,
        )
        appointments = self.appointment.__class__.objects.filter(**opts).exclude(
            id=self.appointment.id
        )
        for appointment in appointments:
            if not appointment.visit:
                appointment.appt_status = INCOMPLETE_APPT
            else:
                if (
                    appointment.crf_metadata_required_exists
                    or appointment.requisition_metadata_required_exists
                ):
                    appointment.appt_status = INCOMPLETE_APPT
                else:
                    appointment.appt_status = COMPLETE_APPT
            appointment.save_base(update_fields=["appt_status"])
            appointment.refresh_from_db()



from django.urls.conf import path
from django.views.generic.base import RedirectView

from .admin_site import edc_appointment_admin
from .views import UnscheduledAppointmentView

app_name = "edc_appointment"

urlpatterns = [
    path(
        "unscheduled_appointment/<subject_identifier>/<visit_schedule_name>"
        "/<schedule_name>/<visit_code>/<timepoint>/<redirect_url>/",
        UnscheduledAppointmentView.as_view(),
        name="unscheduled_appointment_url",
    ),
    path("admin/", edc_appointment_admin.urls),
    path("", RedirectView.as_view(url="/edc_appointment/admin/"), name="home_url"),
]



import arrow


class PreviousAppointmentFormMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._previous_appointment_rdate = None
        self._previous_appointment = None

    @property
    def previous_appointment(self):
        if not self._previous_appointment:
            cleaned_data = self.cleaned_data
            self._previous_appointment = cleaned_data.get(
                "subject_visit"
            ).appointment.previous_by_timepoint
        return self._previous_appointment

    @property
    def previous_appointment_rdate(self):
        """Returns the utc arrow object of appt_datetime of the previous
        appointment

        Usage:
            from django.conf import settings
            rdate.to(settings.TIME_ZONE).date()
            rdate.to(settings.TIME_ZONE).datetime
        """
        if not self._previous_appointment_rdate:
            if self.previous_appointment:
                rdate = arrow.Arrow.fromdatetime(
                    self.previous_appointment.appt_datetime,
                    tzinfo=self.previous_appointment.appt_datetime.tzinfo,
                )
            else:
                rdate = None
            self._previous_appointment_rdate = rdate
        return self._previous_appointment_rdate



from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.http.response import HttpResponseRedirect
from django.urls.base import reverse
from django.utils.safestring import mark_safe
from django.views.generic.base import View

from ..creators import (
    AppointmentInProgressError,
    InvalidParentAppointmentMissingVisitError,
    InvalidParentAppointmentStatusError,
    UnscheduledAppointmentCreator,
    UnscheduledAppointmentError,
)


class UnscheduledAppointmentView(View):

    """A view that creates an unscheduled appointment and redirects to
    the subject dashboard.

    Source Url is usually reversed in the Appointment model wrapper.
    Redirect url name comes in kwargs.
    """

    unscheduled_appointment_cls = UnscheduledAppointmentCreator
    dashboard_template = "subject_dashboard_template"

    def get(self, request, *args, **kwargs):

        try:
            creator = self.unscheduled_appointment_cls(**kwargs)
        except (
            ObjectDoesNotExist,
            UnscheduledAppointmentError,
            InvalidParentAppointmentMissingVisitError,
            InvalidParentAppointmentStatusError,
            AppointmentInProgressError,
        ) as e:
            messages.error(self.request, str(e))
        else:
            messages.success(
                self.request,
                mark_safe(f"Appointment {creator.appointment} was created successfully."),
            )
            messages.warning(
                self.request,
                mark_safe(
                    f"Remember to adjust the appointment date and time on "
                    f"appointment {creator.appointment}."
                ),
            )
        return HttpResponseRedirect(
            reverse(
                self.kwargs.get("redirect_url"),
                kwargs={"subject_identifier": kwargs.get("subject_identifier")},
            )
        )



# Generated by Django 4.0.4 on 2022-05-19 19:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('edc_appointment', '0026_auto_20220501_1548'),
    ]

    operations = [
        migrations.AddField(
            model_name='appointment',
            name='document_status',
            field=models.CharField(choices=[('INCOMPLETE', 'Incomplete (some data pending)'), ('COMPLETE', 'Complete')], default='COMPLETE', help_text='If some data is still pending, flag as incomplete', max_length=25, verbose_name='Document status'),
        ),
        migrations.AddField(
            model_name='historicalappointment',
            name='document_status',
            field=models.CharField(choices=[('INCOMPLETE', 'Incomplete (some data pending)'), ('COMPLETE', 'Complete')], default='COMPLETE', help_text='If some data is still pending, flag as incomplete', max_length=25, verbose_name='Document status'),
        ),
    ]
